# -*- coding: utf-8 -*-
"""
Atomic structure parsers.
"""
import re
import warnings
from abc import abstractmethod
from contextlib import AbstractContextManager, suppress
from functools import lru_cache
from itertools import repeat
from os import PathLike, environ
from pathlib import Path
from platform import system
from string import digits, punctuation
from tempfile import gettempdir
from typing import Any, Iterable, Optional, Tuple, Union
from urllib.error import URLError
from urllib.request import urlretrieve
from warnings import warn

import CifFile as Cif
import numpy as np
import requests
from CifFile import ReadCif
from numpy.linalg import inv

from . import __version__
from .affine import affine_map, transform
from .atom import Atom, frac_coords
from .biological import Helix, Residue, SecondaryStructure, Sheet
from .lattice import Lattice
from .spg_data import HM2Hall, Number2Hall, SymOpsHall

# Temporary directory in which to cache crystal structure files
# downloaded from the internet.
# Note : temporary directory on MacOS systems (Darwin) do not work with gettempdir
#        In this case, we use "/tmp"
STRUCTURE_CACHE = (
    Path("/tmp" if system() == "Darwin" else gettempdir()) / "crystals_cache"
)

MISSING_MP_API_KEY_MESSAGE = """
You have not provided a Materials Project API key. Either provide it, or set it 
as the `MATERIALS_PROJECT_API_KEY` environment variable.
"""


def get_number_with_esd(x: Any) -> Tuple[float, float]:
    """pycifrw's version cannot handle floats, only strings."""
    try:
        return Cif.get_number_with_esd(x)
    except TypeError:
        return Cif.get_number_with_esd(str(x))


class ParseError(IOError):
    """General parsing error type."""

    pass


class AbstractStructureParser(AbstractContextManager):
    """
    Abstract base class for structure parsers. The preferred method
    of using this object is as a context manager.

    Parameters
    ----------
    filename : str or path-like
        Location of the CIF file.
    """

    @abstractmethod
    def __init__(self, *args, **kwargs):
        pass

    @abstractmethod
    def lattice_vectors(self) -> Iterable[np.ndarray]:
        """
        Returns the lattice vectors associated to a structure.

        Returns
        -------
        lv : iterable of ndarrays, shape (3,)
        """
        pass

    def symmetry_operators(self) -> Iterable[np.ndarray]:
        """
        Returns the symmetry operators that map the fractional atomic positions in a
        structure to the crystal *conventional* unit cell.

        Returns
        -------
        sym_ops : iterable of ndarray, shape (4,4)
            Transformation matrices. Since translations and rotation are combined,
            the transformation matrices are 4x4.
        """
        # Default implementation is the trivial transformation
        return [np.eye(4)]

    @abstractmethod
    def atoms(self):
        """
        Asymmetric unit cell. Combine with AbstractStructureParser.symmetry_operators()
        for a full unit cell.

        Returns
        -------
        atoms : iterable of Atom instance
        """
        pass


class PDBParser(AbstractStructureParser):
    """
    Collection of methods that parses Protein DataBank (PDB) files. This object should be used as a context manager.

    Parameters
    ----------
    ID : str
        Protein DataBank identification. The correct .pdb file will be downloaded,
        cached and parsed.
    download_dir : path-like object or None, optional
        Directory where to save the PDB file.
    overwrite : bool, optional
        Whether or not to overwrite files in cache if they exist. If no revision
        number is provided, files will always be overwritten.
    """

    def __init__(
        self, ID: str, download_dir: Optional[PathLike] = None, overwrite: bool = False
    ):
        if download_dir is None:
            download_dir = STRUCTURE_CACHE

        filename = self.download_pdb_file(
            pdb_code=ID, download_dir=download_dir, overwrite=overwrite
        )
        self._handle = open(filename, "r")

    def __exit__(self, *args, **kwargs):
        self._handle.close()

    @staticmethod
    def download_pdb_file(
        pdb_code: str,
        download_dir: Optional[PathLike] = None,
        server: str = "https://files.rcsb.org",
        overwrite: bool = False,
    ) -> Path:
        """
        Retrieves a PDB structure file from a PDB server and
        stores it in a local file tree.

        Parameters
        ----------
        pdf_code : str, len 4
            PDB ID code
        download_dir : path-like object
            Directory where to save the PDB file. Default is a local folder in the current directory
        server : str, optional
            Root address of the server from which to download the PDB file. Default is the main server.
        overwrite : bool, optional
            If True, existing PDB file with the same structure will be overwritten. Default is False.

        Returns
        -------
        file : pathlib.Path
            Pointer to the downloaded file
        """
        if download_dir is None:
            path = STRUCTURE_CACHE
        else:
            path = Path(download_dir)

        path.mkdir(exist_ok=True)

        final_file = path / f"pdb{pdb_code.lower()}.ent"  # (decompressed)

        # Skip download if the file already exists
        if (not overwrite) and (final_file.exists()):
            return final_file

        resp = requests.get(server + f"/download/{pdb_code.upper()}.pdb")
        resp.raise_for_status()

        with open(final_file, "wb") as out:
            out.write(resp.content)

        return Path(final_file)

    @property
    def filename(self) -> str:
        return self._handle.name

    @lru_cache(maxsize=1)
    def lattice_vectors(self) -> Iterable[np.ndarray]:
        """
        Returns the lattice vectors associated to a PDB structure.

        Returns
        -------
        lv : list of ndarrays, shape (3,)

        Raises
        ------
        ParseError
            If the file does not contain a CRYST1 tag.
        """
        self._handle.seek(0)

        for line in filter(lambda l: l.startswith("CRYST1"), self._handle):
            # characters are described in the PDB file content guide.
            a, b, c = float(line[6:15]), float(line[15:24]), float(line[24:33])
            alpha, beta, gamma = (
                float(line[33:40]),
                float(line[40:47]),
                float(line[47:54]),
            )
            break
        else:
            raise ParseError("No CRYST1 line found")

        return Lattice.from_parameters(a, b, c, alpha, beta, gamma).lattice_vectors

    def helices(
        self, ignored: Iterable[str] = ("HOH", "LI1", "SQU")
    ) -> Iterable[Helix]:
        """Returns an iterable of helices present in the protein.

        Parameters
        ----------
        ignored : iterable of str, optional
            3-letter string code for residues to ignore.

        Returns
        -------
        hex : iterable of Helix instances
        """
        residues = self.residues(ignored)
        helices = list()
        self._handle.seek(0)

        # Helices are defined as a list of residues making them up
        # Therefore, we simply filter residues down to the ones in range
        for line in filter(lambda line: line.startswith("HELIX"), self._handle):
            seq_num = int(line[7:10])  # TODO: place in helix
            seq_range = range(int(line[21:25]), int(line[33:37]) + 1)
            helices.append(
                Helix(
                    residues=filter(
                        lambda res: res.sequence_number in seq_range, residues
                    ),
                    sequence_number=seq_num,
                )
            )

        return helices

    def sheets(self, ignored: Iterable[str] = ("HOH", "LI1", "SQU")) -> Iterable[Sheet]:
        """Returns an iterable of sheets present in the protein.

        Parameters
        ----------
        ignored : iterable of str, optional
            3-letter string code for residues to ignore.

        Returns
        -------
        sheets : iterable of Sheet instances
        """
        residues = self.residues(ignored)
        sheets = list()
        self._handle.seek(0)

        # Helices are defined as a list of residues making them up
        # Therefore, we simply filter residues down to the ones in range
        for line in filter(lambda line: line.startswith("SHEET"), self._handle):
            seq_num = int(line[7:10])  # TODO: place in sheet
            seq_range = range(int(line[22:26]), int(line[33:37]) + 1)
            sheets.append(
                Sheet(
                    residues=filter(
                        lambda res: res.sequence_number in seq_range, residues
                    ),
                    sequence_number=seq_num,
                )
            )

        return sheets

    def secondary_structures(
        self, ignored: Iterable[str] = ("HOH", "LI1", "SQU")
    ) -> Iterable[SecondaryStructure]:
        """Iterable of all secondary structures present in this file.

        Parameters
        ----------
        ignored : iterable of str, optional
            3-letter string code for residues to ignore.

        Returns
        -------
        structures : iterable of SecondaryStructure instances
        """
        return list(self.sheets(ignored=ignored)) + list(self.helices(ignored=ignored))

    def residues(
        self, ignored: Iterable[str] = ("HOH", "LI1", "SQU")
    ) -> Iterable[Residue]:
        """
        Iterable of residues present in the structure.

        Parameters
        ----------
        ignored : iterable of str, optional
            3-letter string code for residues to ignore.

        Returns
        -------
        res : iterable of Residue instances
        """
        # Lattice vectors have to be determined first because
        # the file pointer is moved
        lattice_vectors = self.lattice_vectors()

        # Filter lines with start with ATM or HETATM
        is_atom_line = lambda l: l.startswith(("ATOM", "HETATM"))

        # ``residues`` is a dictionary mapping between (sequence numbers, name)
        # and an iterable of ``Atom``. When we have collected all atoms, we then create
        # a ``Residue`` for each sequence number
        residues = dict()

        self._handle.seek(0)
        for line in filter(is_atom_line, self._handle):
            residue_name = str(line[17:20]).replace(" ", "")
            if residue_name in ignored:
                continue

            residue_seq = int(line[22:26])
            if (residue_seq, residue_name) not in residues:
                residues[(residue_seq, residue_name)] = list()

            # TODO: include Atom ID record in Atom objects
            identification = str(line[12:16]).replace(" ", "")
            x, y, z = float(line[30:38]), float(line[38:46]), float(line[46:54])
            coords_fractional = frac_coords([x, y, z], lattice_vectors)
            element = str(line[76:78]).replace(" ", "")

            try:
                occupancy = float(line[54:60])
            except ValueError:
                occupancy = None

            residues[(residue_seq, residue_name)].append(
                Atom(element=element, coords=coords_fractional, occupancy=occupancy)
            )

        if not residues:
            raise ParseError(f"No residues found in {self.filename}")

        structures = list()
        for (seq_number, name), atoms in residues.items():
            structures.append(
                Residue(atoms=atoms, name=name, sequence_number=seq_number)
            )
        return structures

    def atoms(self) -> Iterable[Atom]:
        """
        Returns a list of atoms associated with a PDB structure in fractional coordinates.
        These atoms form the asymmetric unit cell.

        Returns
        -------
        atoms: iterable of Atom
        """
        # Lattice vectors have to be determined first because
        # the file pointer is moved
        lattice_vectors = self.lattice_vectors()

        is_atom_line = lambda l: l.startswith(("ATOM", "HETATM"))

        self._handle.seek(0)
        atoms = list()
        for line in filter(is_atom_line, self._handle):
            element = str(line[76:78]).replace(" ", "").title()

            x, y, z = float(line[30:38]), float(line[38:46]), float(line[46:54])
            fractional_coordinates = frac_coords(np.array([x, y, z]), lattice_vectors)

            try:
                occupancy = float(line[54:60])
            except ValueError:
                occupancy = None

            atoms.append(
                Atom(
                    element=element, coords=fractional_coordinates, occupancy=occupancy
                )
            )
        return atoms

    def symmetry_operators(self) -> Iterable[np.ndarray]:
        """
        Returns the symmetry operators that map the atomic positions in a
        PDB file to the crystal unit cell.

        Returns
        -------
        sym_ops : iterable of `~numpy.ndarray`, shape (4,4)
            Transformation matrices. Since translations and rotation are combined,
            the transformation matrices are 4x4.
        """
        self._handle.seek(0)

        # This is clunky af
        sym_ops = dict()
        for line in filter(
            lambda l: l.startswith("REMARK 290") and ("SMTRY" in l), self._handle
        ):
            op_num = line[22:23]
            if op_num not in sym_ops:
                sym_ops[op_num] = {"rotation": list(), "translation": list()}

            r1, r2, r3, t = np.fromstring(line[23:], dtype=float, count=4, sep=" ")
            sym_ops[op_num]["rotation"].append([r1, r2, r3])
            sym_ops[op_num]["translation"].append(t)

        if not sym_ops:
            raise ParseError(
                f"No symmetry could be parsed from file {self._handle.filename}"
            )

        operators = list()
        for op in sym_ops.values():
            mat = np.eye(4, dtype=float)
            mat[:3, :3] = np.array(op["rotation"])
            mat[:3, 3] = np.array(op["translation"])
            operators.append(mat)

        return operators


class CIFParser(AbstractStructureParser):
    """
    Collection of methods that parses CIF files based on cif2cell. The preferred method
    of using this object is as a context manager.

    Parameters
    ----------
    filename : str or path-like
        Location of the CIF file.
    """

    def __init__(self, filename: PathLike, **kwargs):
        # ReadCIF would get confused between local files and URLs
        # Therefore, more clear to pass an open file
        self._handle = open(filename, mode="r")
        self.file = ReadCif(self._handle, **kwargs)

    def __exit__(self, *args, **kwargs):
        self._handle.close()

    @property
    def filename(self) -> str:
        return self._handle.name

    @staticmethod
    def sym_ops_from_equiv(equiv_site: Union[str, Iterable[str]]) -> np.ndarray:
        """Parse a symmetry operator from an equivalent-site representation

        Parameters
        ----------
        equiv_site : str or iterable of strings
            Either comma-separated string e.g. "+y, +x, -z + 1/2" or an
            iterable of the comma-separated values, e.g. ["+y", "+x", "-z + 1/2"]

        Returns
        -------
        sym_ops : ndarray, shape (4,4)
            Symmetry operator as a 4x4 affine transformation on the FRACTIONAL
            coordinates.
        """
        symmetry = np.zeros((3, 3))
        translation = np.zeros((3,))

        if isinstance(equiv_site, str):
            equiv_site = equiv_site.split(",")

        equiv_site = tuple(map(lambda s: s.strip().lower(), equiv_site))
        for j in range(3):
            xyz = equiv_site[j].replace("+", " +").replace("-", " -").split()
            for i in xyz:
                if i.strip("+-") == "x":
                    symmetry[0, j] = float(i.strip("x") + "1")
                elif i.strip("+-") == "y":
                    symmetry[1, j] = float(i.strip("y") + "1")
                elif i.strip("+-") == "z":
                    symmetry[2, j] = float(i.strip("z") + "1")

                if i.strip("+-xyz") != "":
                    translation[j] = eval(i)

        symmetry[:] = np.transpose(symmetry)

        # Combination of transform and translation into a single matrix
        # is done in a 4x4 affine transform
        symmetry_operation = affine_map(symmetry)
        symmetry_operation[:3, 3] = translation
        return symmetry_operation

    @property
    def structure_block(self):
        """Retrieve which CIF block has the appropriate structural information"""
        blocks = (self.file[key] for key in self.file.keys())
        for block in blocks:
            try:
                _, _ = get_number_with_esd(block["_cell_length_a"])
            except KeyError:
                continue
            else:
                return block

    @lru_cache(maxsize=1)
    def hall_symbol(self) -> str:
        """Returns the Hall symbol"""
        block = self.structure_block

        hall_symbol = block.get("_symmetry_space_group_name_Hall") or block.get(
            "_space_group_name_Hall"
        )

        # In some rare cases, the given hall symbol in the file isn't standard,
        # otherwise it would be a key in SymOpsHall
        # Then, it is preferable to infer the conventional hall symbol from other info
        if (hall_symbol is None) or (hall_symbol not in SymOpsHall):
            h_m_symbol = block.get("_symmetry_space_group_name_H-M") or block.get(
                "_space_group_name_H-M_alt"
            )

            if h_m_symbol is not None:
                h_m_symbol = re.sub(r"\s+", "", h_m_symbol)
                with suppress(
                    KeyError
                ):  # Symbol could be meaningless, e.g. h_m_symbol = '?' (True story)
                    hall_symbol = HM2Hall[h_m_symbol]

        # Again, if hall_symbol is still missing OR invalid
        if (hall_symbol is None) or (hall_symbol not in SymOpsHall):
            table_number = block.get("_symmetry_Int_Tables_number") or block.get(
                "_space_group_IT_number"
            )

            if table_number is not None:
                hall_symbol = Number2Hall.get(int(table_number))

        if hall_symbol is None:
            raise ParseError("Hall symbol could not be inferred")

        if hall_symbol[0] == "-":
            hall_symbol = "-" + hall_symbol[1].upper() + hall_symbol[2:].lower()
        else:
            hall_symbol = hall_symbol[0].upper() + hall_symbol[1:].lower()

        return hall_symbol

    @lru_cache(maxsize=1)
    def lattice_parameters(self) -> Tuple[float, float, float, float, float, float]:
        """
        Returns the lattice parameters associated to a CIF structure.

        Returns
        ----------
        a, b, c : float
            Lengths of lattice vectors [Angstroms]
        alpha, beta, gamma : float
            Angles of lattice vectors [degrees].
        """
        block = self.structure_block

        try:
            a_with_err = block["_cell_length_a"]
        except KeyError:
            raise ParseError(f"No lattice information is present in {self.filename}")

        # In case where b and c are not listed, we use the value of a
        a, _ = get_number_with_esd(a_with_err)
        b, _ = get_number_with_esd(block.get("_cell_length_b", a_with_err))
        c, _ = get_number_with_esd(block.get("_cell_length_c", a_with_err))

        alpha, _ = get_number_with_esd(block["_cell_angle_alpha"])
        beta, _ = get_number_with_esd(block["_cell_angle_beta"])
        gamma, _ = get_number_with_esd(block["_cell_angle_gamma"])

        return a, b, c, alpha, beta, gamma

    @lru_cache(maxsize=1)
    def lattice_vectors(self) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """
        Returns the lattice vectors associated to a CIF structure.

        Returns
        -------
        lv : list of ndarrays, shape (3,)
        """
        return Lattice.from_parameters(*self.lattice_parameters()).lattice_vectors

    def symmetry_operators(self) -> Iterable[np.ndarray]:
        """
        Returns the symmetry operators that map the fractional atomic positions in a
        CIF file to the crystal *conventional* unit cell.

        Returns
        -------
        sym_ops : iterable of ndarray, shape (4,4)
            Transformation matrices. Since translations and rotation are combined,
            the transformation matrices are 4x4.
        """
        block = self.structure_block

        equivalent_sites_str = None
        for tag in ["_symmetry_equiv_pos_as_xyz", "_space_group_symop_operation_xyz"]:
            with suppress(KeyError):
                equivalent_sites_str = block.GetLoop(tag).get(tag)

        # P1 space group only has a single equivalent site
        if isinstance(equivalent_sites_str, str):
            equivalent_sites_str = [equivalent_sites_str]

        with suppress(ParseError):
            if not equivalent_sites_str:
                equivalent_sites_str = SymOpsHall[self.hall_symbol()]
            elif len(equivalent_sites_str) != len(SymOpsHall[self.hall_symbol()]):
                warnings.warn(
                    f"The number of equivalent sites is not in line with the database. The file {self.filename} might be incomplete"
                )

        return list(map(self.sym_ops_from_equiv, equivalent_sites_str))

    def atoms(self) -> Iterable[Atom]:
        """
        Asymmetric unit cell. Combine with CIFParser.symmetry_operators() for a full unit cell.

        Returns
        -------
        atoms : iterable of Atom instance
        """
        block = self.structure_block

        try:
            tmpdata = block.GetLoop("_atom_site_fract_x")
            cartesian = False
        except:
            try:
                tmpdata = block.GetLoop("_atom_site_Cartn_x")
                cartesian = True
            except:
                raise ParseError("Atomic positions could not be found or inferred.")

            t11 = block.get("_atom_sites_Cartn_tran_matrix_11")
            t12 = block.get("_atom_sites_Cartn_tran_matrix_12")
            t13 = block.get("_atom_sites_Cartn_tran_matrix_13")
            t21 = block.get("_atom_sites_Cartn_tran_matrix_21")
            t22 = block.get("_atom_sites_Cartn_tran_matrix_22")
            t23 = block.get("_atom_sites_Cartn_tran_matrix_23")
            t31 = block.get("_atom_sites_Cartn_tran_matrix_13")
            t32 = block.get("_atom_sites_Cartn_tran_matrix_23")
            t33 = block.get("_atom_sites_Cartn_tran_matrix_33")
            cart_trans_matrix_inv = np.array(
                [
                    [float(t11), float(t12), float(t13)],
                    [float(t21), float(t22), float(t23)],
                    [float(t31), float(t32), float(t33)],
                ]
            )
            cart_trans_matrix = inv(cart_trans_matrix_inv)

            if not all([t11, t12, t13, t21, t22, t23, t31, t32, t33]):
                raise ParseError(
                    "Cartesian coordinates in CIF but no transformation matrix given"
                )

        if cartesian:
            xs = tmpdata.get("_atom_site_Cartn_x")
            ys = tmpdata.get("_atom_site_Cartn_y")
            zs = tmpdata.get("_atom_site_Cartn_z")
        else:
            xs = tmpdata.get("_atom_site_fract_x")
            ys = tmpdata.get("_atom_site_fract_y")
            zs = tmpdata.get("_atom_site_fract_z")
        # TODO: handle wildcards like '?', '.' in xs, ys, zs

        elements = tmpdata.get("_atom_site_type_symbol")
        if not elements:
            elements = tmpdata.get("_atom_site_label")
            if not elements:
                raise ParseError("Atom symbols could not be found or inferred.")
        elements = map(lambda s: s.strip(punctuation + digits).title(), elements)

        occupancies = tmpdata.get("_atom_site_occupancy")
        if not occupancies:
            # len(xs) == number of atoms. Therefore, good upper bound for `repeat`, which otherwise
            # might produce an infinitely long iterable
            occupancies = repeat(1.0, len(xs))
        occupancies = map(
            lambda x: get_number_with_esd(x)[0], occupancies
        )  # See issue #7

        atoms = list()
        for e, x, y, z, occ in zip(elements, xs, ys, zs, occupancies):
            coords = np.array(
                [
                    get_number_with_esd(x)[0],
                    get_number_with_esd(y)[0],
                    get_number_with_esd(z)[0],
                ]
            )

            # We normalize atom position to be within the unit cell
            # Therefore we need the fractional coordinates
            if cartesian:
                coords = transform(cart_trans_matrix, coords)
                coords[:] = frac_coords(coords, self.lattice_vectors())

            atoms.append(Atom(element=e, coords=np.mod(coords, 1), occupancy=occ))

        return atoms


class CODParser(CIFParser):
    """
    Collection of methods that parses CIF files retrieved from the Crystallography Open Database.
    The preferred method of using this object is as a context manager.

    Parameters
    ----------
    num : int
        COD identification number.
    revision : int or None, optional
        Revision number. If None (default), the latest revision is used.
    download_dir : path-like object or None, optional
        Directory where to save the CIF file.
    overwrite : bool, optional
        Whether or not to overwrite files in cache if they exist. If no revision
        number is provided, files will always be overwritten.

    Raises
    ------
    RuntimeError : If the file could not be downloaded from any of the mirrors.
    """

    # Database mirrors are made available
    # see http://wiki.crystallography.net/codmirrors/
    mirrors = (
        "http://www.crystallography.net/cod/",
        "http://cod.ibt.lt/cod",
        "http://qiserver.ugr.es/cod/",
    )

    def __init__(
        self,
        num: int,
        revision: Optional[int] = None,
        download_dir: Optional[PathLike] = None,
        overwrite: bool = False,
        **kwargs,
    ):
        if download_dir is None:
            download_dir = STRUCTURE_CACHE

        super().__init__(
            filename=self.download_cif(download_dir, num, revision, overwrite), **kwargs
        )

    @classmethod
    def download_cif(
        cls,
        download_dir: PathLike,
        num: int,
        revision: Optional[int] = None,
        overwrite: bool = False,
    ) -> Path:
        """
        Download a CIF file from the Crystallography Open Database.

        Parameters
        ----------
        download_dir : path-like object
            Directory where to save the CIF file.
        num : int
            COD identification number.
        revision : int or None, optional
            Revision number. If None (default), the latest revision is used.
        overwrite : bool, optional
            Whether or not to overwrite files in cache if they exist. If no revision
            number is provided, files will always be overwritten.

        Returns
        -------
        path : pathlib.Path
            Path to the downloaded file.

        Raises
        ------
        RuntimeError : If the file could not be downloaded from any of the mirrors.

        Notes
        -----
        This function will try three download mirrors. Warnings will be emitted in case
        the file cannot be found in a mirror.
        """
        download_dir = Path(download_dir)
        download_dir.mkdir(exist_ok=True)

        if revision is None:
            # We can never be sure of what the latest revision is
            # Therefore, to be safe, we overwrite
            overwrite = True

            # If latest revision, url should be e.g. (mirror)/1023891.cif
            url_suffix = f"{num}.cif"
            base = f"{num}.cif"
        else:
            # For revision, url should be e.g. (mirror)/9812812.cif@98181234
            url_suffix = str(num) + ".cif@" + str(revision)
            base = f"{num}-{revision}.cif"

        download_path = download_dir / base

        if download_path.exists() and (not overwrite):
            return download_path

        for index, mirror in enumerate(cls.mirrors, start=1):
            url = mirror + url_suffix
            try:
                urlretrieve(url, download_path)
            except URLError as e:
                warn(f"The file {url} could not be downloaded because: {e.reason}")

                # If this is the last mirror, then the file could not
                # be downloaded at all
                if index == len(cls.mirrors):
                    raise RuntimeError(
                        f"Crystallography Open Database ID {num} could not be downloaded from any mirror."
                    ) from None

                continue
            else:
                break

        return download_path


class MPJParser(CIFParser):
    """
    Collection of methods that parses CIF files retrieved from the Materials Project.
    The preferred method of using this object is as a context manager.

    Parameters
    ----------
    query : str
        The query can be a Materials Project material id (e.g., `"mp-1234"`), a
        formula, e.g. (`"Fe2O3"`), or a chemical system ("-" separated list of elemments,
        e.g., `"Li-Fe-O"`).
    api_key : str or None, optional
        An API key for accessing the Materials Project REST interface.
        Please obtain your API key at https://www.materialsproject.org/dashboard.
        If `None` (default), ``crystals`` will look for your API key in the
        `MATERIALS_PROJECT_API_KEY` environment variable.
    download_dir : path-like object or None, optional
        Directory where to save the CIF file. This is used for caching.
    overwrite : bool, optional
        Whether or not to overwrite files in cache if they exist. If True,
        a new file will be downloaded, possibly overwriting previously-downloaded file.

    Raises
    ------
    ValueError : if `api_key` is None and it could not be found in the environment variables.
    """

    def __init__(
        self,
        query: str,
        api_key: Optional[str] = None,
        download_dir: Optional[PathLike] = None,
        overwrite: bool = False,
        **kwargs,
    ):
        if download_dir is None:
            download_dir = STRUCTURE_CACHE

        if api_key is None:
            api_key = environ.get("MATERIALS_PROJECT_API_KEY", None)

            if api_key is None:
                raise ValueError(MISSING_MP_API_KEY_MESSAGE)
        super().__init__(
            filename=self.download_cif(
                api_key=api_key,
                query=query,
                download_dir=download_dir,
                overwrite=overwrite,
            ),
            **kwargs,
        )

    @property
    def material_id(self):
        """Returns the Materials Project material ID from this file."""
        # A comment of the form "Material ID: xxxxxxx" will have been inserted in the first
        # line of the file after download.
        self._handle.seek(0)
        firstline = next(iter(self._handle))
        *_, material_id = firstline.split(":")
        return material_id.strip()

    def download_cif(
        self,
        query: str,
        api_key: Optional[str] = None,
        download_dir: Optional[PathLike] = None,
        overwrite: bool = False,
    ) -> Path:
        """
        Download a CIF file from the Materials Project Database.

        Parameters
        ----------
        api_key : str
            An API key for accessing the Materials Project REST interface.
            Please obtain your API key at https://www.materialsproject.org/dashboard.
        query : str
            The query can be a Materials Project material id (e.g., `"mp-1234"`), a
            formula, e.g. (`"Fe2O3"`), or a chemical system ("-" separated list of elemments,
            e.g., `"Li-Fe-O"`).
        download_dir : path-like object or None, optional
            Directory where to save the CIF file. This is used for caching.
        overwrite : bool, optional
            Whether or not to overwrite files in cache if they exist. If True,
            a new file will be downloaded, possibly overwriting previously-downloaded file.

        Returns
        -------
        path : pathlib.Path
            Path to the downloaded file.

        Raises
        ------
        ConnectionError : If the file could not be downloaded.
        """
        download_dir = Path(download_dir)
        download_dir.mkdir(exist_ok=True)
        target_filename = download_dir / f"{query}.cif"

        if target_filename.exists() and (not overwrite):
            return target_filename

        endpoint = f"https://materialsproject.org/rest/v2/materials/{query}/vasp/cif"
        headers = {
            "x-api-key": api_key,
        }

        with requests.get(endpoint, headers=headers) as response:
            if response.status_code == 403:  # Forbidden access
                raise ConnectionError(
                    "Materials Project API key is not authorized to do this operation."
                    / f"status code {response.status_code}"
                )
            elif response.status_code != 200:
                raise ConnectionError(
                    f"Would not connect: status code {response.status_code}"
                )

            body = response.json()["response"][0]
            material_id = body["material_id"]

            with open(target_filename, "w") as f:
                # We write the material ID as a comment at the top of the CIF file
                # so we can inform the user in the Crystal.source property
                f.write(f"# Material ID: {material_id}\n")
                f.write(body["cif"])

        return target_filename


class PWSCFParser(AbstractStructureParser):
    """
    Collection of methods that parses output files from the Plane-Wave Self-Consistent
    Field (PWSCF) program, part of the Quantum Espresso suite.

    The preferred method of using this object is as a context manager.

    Parameters
    ----------
    filename : str or path-like
        Location of the CIF file.
    """

    # Regular expression pattern for a 3-vector
    # These are numbers separated by whitespace, in parentheses
    # Example:
    #   (  -0.0008701   0.5704561   0.4409210  )
    _vector_pattern = r"[(]\s* (?P<x1>[-]?[0-9]*\.[0-9]+\s*) (?P<x2>[-]?[0-9]*\.[0-9]+\s*) (?P<x3>[-]?[0-9]*\.[0-9]+\s*) [)]"

    # Conversion factor from bohr radius to angstroms
    _bohr_to_angs = 0.529_177_249

    def __init__(self, filename: PathLike, **kwargs):
        self.filename = filename

        with open(filename, mode="r") as f:
            self._filecontent = f.read()

    def __exit__(self, *args, **kwargs):
        pass

    @property
    def alat(self) -> float:
        """Get the lattice parameter [Bohr radius]"""
        match = re.search(
            r"\s*(lattice parameter [(]alat[)])\s*=\s*(?P<alat>\d+[.]\d+)\s*(a.u.)",
            self._filecontent,
        )
        if not match:
            raise ParseError(
                f"Lattice parameter from {self.filename} could not be parsed."
            )

        return float(match.group("alat"))

    @property
    def natoms(self) -> int:
        """Number of atoms defined per cell"""
        match = re.search(
            r"(\s*number of atoms/cell\s*)[=]\s*(?P<natoms>\d+)", self._filecontent
        )
        if not match:
            raise ParseError(
                f"Lattice parameter from {self.filename} could not be parsed."
            )

        return int(match.group("natoms"))

    @property
    def celldm(self) -> Tuple[float, float, float, float, float, float]:
        """Crystallographic constants as defined by INPUT_PW. They are, in order:

        * a,b,c in ANGSTROM

        * cosAB = cosine of the angle between axis a and b (gamma)

        * cosAC = cosine of the angle between axis a and c (beta)

        * cosBC = cosine of the angle between axis b and c (alpha)

        The constants are placed in a dictionary, indexed between 1 and 6.
        """
        params = dict()
        for index in (1, 2, 3, 4, 5, 6):
            pattern = (
                r"(celldm[(]"
                + str(index)
                + r"[)]=\s*)(?P<value>[-]?[0-9]*\.[0-9]+)(\s*)"
            )
            match = re.search(pattern, self._filecontent).group("value")
            params[index] = float(match)

        return params

    def lattice_vectors_alat(self) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """
        Returns the lattice vectors associated to a structure. These lattice vectors are in
        units of `lattice parameters` [alat].

        Returns
        -------
        lv : iterable of ndarrays, shape (3,)
        """
        a1 = re.search(
            r"(\s*a[(]1[)]\s*=\s*)" + self._vector_pattern, self._filecontent
        ).group("x1", "x2", "x3")
        a2 = re.search(
            r"(\s*a[(]2[)]\s*=\s*)" + self._vector_pattern, self._filecontent
        ).group("x1", "x2", "x3")
        a3 = re.search(
            r"(\s*a[(]3[)]\s*=\s*)" + self._vector_pattern, self._filecontent
        ).group("x1", "x2", "x3")

        return tuple(np.array(tuple(map(float, a))) for a in (a1, a2, a3))

    def reciprocal_vectors_alat(self) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """
        Returns the reciprocal lattice vectors associated to a structure, in units of :math:`2 \\pi / alat`.

        Returns
        -------
        lv : iterable of ndarrays, shape (3,)
        """
        b1 = re.search(
            r"(\s*b[(]1[)]\s*=\s*)" + self._vector_pattern, self._filecontent
        ).group("x1", "x2", "x3")
        b2 = re.search(
            r"(\s*b[(]2[)]\s*=\s*)" + self._vector_pattern, self._filecontent
        ).group("x1", "x2", "x3")
        b3 = re.search(
            r"(\s*b[(]3[)]\s*=\s*)" + self._vector_pattern, self._filecontent
        ).group("x1", "x2", "x3")

        return tuple(np.array(tuple(map(float, b))) for b in (b1, b2, b3))

    def lattice_vectors(self) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """
        Returns the lattice vectors associated to a structure [:math:`Å`].

        Returns
        -------
        lv : iterable of ndarrays, shape (3,)
        """
        vectors = np.array(self.lattice_vectors_alat())
        scale = self._bohr_to_angs * self.alat * np.eye(3)
        return tuple(map(np.squeeze, np.vsplit(scale @ vectors, 3)))

    def atoms(self) -> Iterable[Atom]:
        """
        Asymmetric unit cell.

        Returns
        -------
        atoms : iterable of Atom instance
        """
        atoms = list()
        for index in range(1, self.natoms + 1):
            pattern = (
                r"\s*"
                + str(index)
                + r"\s*(?P<element>[A-Z][a-z]?) (\s* tau[(]\s*(?P<atm_index>\d+)[)]\s*=\s*)"
                + self._vector_pattern
            )
            match = re.search(pattern, self._filecontent)
            coords = np.asarray(tuple(map(float, match.group(4, 5, 6))))

            # Convert coordinates in fractional coordinates
            # Note that the atomic coordinates are not quite real-space, but rather
            # real space in units of lattice_parameter (alat) in bohr radius
            atoms.append(
                Atom(
                    element=match.group("element"),
                    coords=frac_coords(coords, self.lattice_vectors_alat()),
                    tag=int(match.group("atm_index")),
                )
            )

        return atoms


class POSCARParser(AbstractStructureParser):
    """
    Collection of methods that parses POSCAR output files from the Vienna Ab initio
    Simulation Package (VASP) suite.

    The preferred method of using this object is as a context manager.

    Parameters
    ----------
    filename : str or path-like
        Location of the POSCAR file.
    """

    def __init__(self, filename: PathLike, **kwargs):
        self.filename = filename

        with open(filename, mode="r") as f:
            next(f)

            scaling_factor = float(next(f))

            self._lattice_vectors = (
                np.array([next(f).split() for _ in range(3)]).astype(float)
                * scaling_factor
            )
            self._atom_types = list(
                zip(
                    next(f).strip().split(),
                    map(int, next(f).strip().split()),
                )
            )
            self._atoms = []
            flag = next(f)

            if flag.startswith("S"):
                raise NotImplementedError(
                    "Selective dynamics tag in POSCAR files are not supported."
                )
            elif flag[0] in ["C", "c", "K", "k"]:
                for element, nat in self._atom_types:
                    for _ in range(nat):
                        coords = np.array(next(f).strip().split()[:3]).astype(float)
                        coords *= scaling_factor
                        coords = coords @ np.linalg.inv(self._lattice_vectors)
                        self._atoms.append(Atom(element, coords))
            else:
                for element, nat in self._atom_types:
                    for _ in range(nat):
                        coords = np.array(next(f).strip().split()[:3]).astype(float)
                        self._atoms.append(Atom(element, coords))

    def __exit__(self, *args, **kwargs):
        pass

    def atoms(self) -> Iterable[Atom]:
        """
        Unit cell.

        Returns
        -------
        atoms : iterable of Atom instance
        """
        return self._atoms

    def lattice_vectors(self) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """
        Returns the lattice vectors associated to a POSCAR structure.

        Returns
        -------
        lv : list of ndarrays, shape (3,)
        """
        return self._lattice_vectors



# -*- coding: utf-8 -*-
"""
This package allows for manipulation and modelling of atomic structures in crystalline form.
"""
__author__ = "Laurent P. René de Cotret"
__email__ = "laurent.decotret@outlook.com"
__license__ = "GPLv3"
__version__ = "1.7.0"

from .atom import (
    Atom,
    ElectronicStructure,
    Element,
    Orbital,
    distance_cartesian,
    distance_fractional,
    frac_coords,
    is_element,
    real_coords,
)
from .atom_data import (
    ELEM_TO_MAGMOM,
    ELEM_TO_MASS,
    ELEM_TO_NAME,
    ELEM_TO_NUM,
    NUM_TO_ELEM,
)
from .base import AtomicStructure
from .biological import Helix, Residue, Sheet
from .crystal import (
    CenteringType,
    Crystal,
    Supercell,
    symmetry_expansion,
    symmetry_reduction,
)
from .indexing import IndexingError, index_dirax
from .lattice import Lattice, LatticeSystem, lattice_system
from .parsers import CIFParser, CODParser, MPJParser, ParseError, PDBParser, PWSCFParser
from .writers import ase_atoms



# -*- coding: utf-8 -*-
"""
Linear algebra operations and helpers.

Inspired by Christoph Gohlke's transformation.py <http://www.lfd.uci.edu/~gohlke/>

This module is not directly exported by the `crystals` library. Use it with caution.
"""

import math
from typing import Tuple

import numpy as np
from numpy.typing import ArrayLike

Basis = Tuple[ArrayLike, ArrayLike, ArrayLike]

# standard basis
e1, e2, e3 = np.eye(3)


def affine_map(array: ArrayLike) -> np.ndarray:
    """
    Extends 3x3 transform matrices to 4x4, i.e. general affine transforms.

    Parameters
    ----------
    array : array-like, shape {(3,3), (4,4)}
        Transformation matrix. If shape = (4,4), returned intact.

    Returns
    -------
    extended : ndarray, shape (4,4)
        Extended array

    Raises
    ------
    ValueError : If the transformation matrix is neither 3x3 or 4x4
    """
    array = np.asarray(array)
    if array.shape == (4, 4):  # Already the right shape
        return array
    elif array.shape == (3, 3):
        extended_matrix = np.zeros(shape=(4, 4), dtype=array.dtype)
        extended_matrix[-1, -1] = 1
        extended_matrix[:3, :3] = array
        return extended_matrix
    else:
        raise ValueError(
            "Array shape not 3x3 or 4x4, and thus is not a transformation matrix."
        )


def transform(matrix: ArrayLike, array: ArrayLike) -> np.ndarray:
    """
    Applies a matrix transform on an array.

    Parameters
    ----------
    matrix : array-like, shape {(3,3), (4,4)}
        Transformation matrix.
    array : array-like, shape {(3,), (3,3), (4,4)}
        Array to be transformed. Either a 1x3 vector, or a transformation
        matrix in 3x3 or 4x4 shape.

    Returns
    -------
    transformed : ndarray
        Transformed array, either a 1D vector or a 4x4 transformation matrix

    Raises
    ------
    ValueError : If the transformation matrix is neither 3x3 or 4x4
    """
    array = np.asarray(array)
    matrix = np.asarray(matrix)

    if matrix.shape not in [(3, 3), (4, 4)]:
        raise ValueError(
            f"Input matrix is neither a 3x3 or 4x4 matrix, but \
						  rather of shape {matrix.shape}."
        )

    matrix = affine_map(matrix)
    # Case of a vector (e.g. position vector):
    if array.ndim == 1:
        extended_vector = np.array([0, 0, 0, 1], dtype=array.dtype)
        extended_vector[:3] = array
        return np.dot(matrix, extended_vector)[:3]
    else:
        array = affine_map(array)
    return np.dot(matrix, array)


def translation_matrix(direction: ArrayLike) -> np.ndarray:
    """
    Return matrix to translate by direction vector.

    Parameters
    ----------
    direction : array_like, shape (3,)

    Returns
    -------
    translation : `~numpy.ndarray`, shape (4,4)
        4x4 translation matrix.
    """
    matrix = np.eye(4)
    matrix[:3, 3] = np.asarray(direction)[:3]
    return matrix


def change_of_basis(basis1: Basis, basis2: Basis = (e1, e2, e3)) -> np.ndarray:
    """
    Returns the matrix transforms vectors expressed in one basis,
    to vectors expressed in another basis.

    Parameters
    ----------
    basis1 : list of array_like, shape (3,)
        First basis
    basis2 : list of array_like, shape (3,), optional
        Second basis. By default, this is the standard basis

    Returns
    -------
    cob : `~numpy.ndarray`, shape (3,3)
        Change-of-basis matrix.
    """
    # Calculate the transform that goes from basis 1 to standard basis
    basis1 = [np.asarray(vector).reshape(3, 1) for vector in basis1]
    basis1_to_standard = np.hstack(tuple(basis1))

    # Calculate the transform that goes from standard basis to basis2
    basis2 = [np.asarray(vector).reshape(3, 1) for vector in basis2]
    standard_to_basis2 = np.linalg.inv(np.hstack(tuple(basis2)))

    return np.dot(standard_to_basis2, basis1_to_standard)


def is_basis(basis: Basis) -> bool:
    """
    Returns true if the set of vectors forms a basis. This is done by checking
    whether basis vectors are independent via an eigenvalue calculation.

    Parameters
    ----------
    basis : list of array-like, shape (3,)

    Returns
    -------
    out : bool
        Whether or not the basis is valid.
    """
    return 0 not in np.linalg.eigvals(np.asarray(basis))


def is_rotation_matrix(matrix: ArrayLike) -> bool:
    """
    Checks whether a matrix is orthogonal with unit determinant (1 or -1), properties
    of rotation matrices.

    Parameters
    ----------
    matrix : array-like, shape {(3,3), (4,4)}
        Rotation matrix candidate. If (4,4) matrix is provided,
        only the top-left block matrix of (3,) is checked

    Returns
    -------
    result : bool
        If True, input could be a rotation matrix.
    """
    # TODO: is this necessary? should a composite transformation
    # 		of translation and rotation return True?
    # if matrix.shape == (4,4):
    #    matrix = matrix[:3,:3]
    matrix = np.asarray(matrix)
    is_orthogonal = np.allclose(np.linalg.inv(matrix), np.transpose(matrix))
    unit_determinant = np.allclose(abs(np.linalg.det(matrix)), 1)
    return is_orthogonal and unit_determinant


def rotation_matrix(angle: float, axis: ArrayLike = (0, 0, 1)) -> np.ndarray:
    """
    Return matrix to rotate about axis defined by direction around the origin [0,0,0].

    Parameters
    ----------
    angle : float
        Rotation angle [rad]
    axis : array-like of length 3
        Axis about which to rotate

    Returns
    -------
    matrix : `~numpy.ndarray`, shape (3,3)
        Rotation matrix.

    See also
    --------
    translation_rotation_matrix

    Notes
    -----
    To combine rotation and translations, see
    http://www.euclideanspace.com/maths/geometry/affine/matrix4x4/index.htm
    """
    sina, cosa = math.sin(angle), math.cos(angle)

    # Make sure direction is a numpy vector of unit length
    direction = np.asarray(axis)
    direction = direction / np.linalg.norm(direction)

    # rotation matrix around unit vector
    R = np.diag([cosa, cosa, cosa])
    R += np.outer(direction, direction) * (1.0 - cosa)
    direction *= sina
    R += np.array(
        [
            [0.0, -direction[2], direction[1]],
            [direction[2], 0.0, -direction[0]],
            [-direction[1], direction[0], 0.0],
        ]
    )

    return R


def translation_rotation_matrix(
    angle: float, axis: ArrayLike, translation: ArrayLike
) -> np.ndarray:
    """
    Returns a 4x4 matrix that includes a rotation and a translation.

    Parameters
    ----------
    angle : float
        Rotation angle [rad]
    axis : array-like of length 3
        Axis about which to rotate
    translation : array_like, shape (3,)
        Translation vector

    Returns
    -------
    matrix : `~numpy.ndarray`, shape (4,4)
        Affine transform matrix.
    """
    rmat = affine_map(rotation_matrix(angle=angle, axis=axis))
    rmat[:3, 3] = np.asarray(translation)
    return rmat


def change_basis_mesh(
    xx: np.ndarray, yy: np.ndarray, zz: np.ndarray, basis1: Basis, basis2: Basis
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    Changes the basis of meshgrid arrays.

    Parameters
    ----------
    xx, yy, zz : ndarrays
        Arrays of equal shape, such as produced by numpy.meshgrid.
    basis1 : list of ndarrays, shape(3,)
        Basis of the mesh
    basis2 : list of ndarrays, shape(3,)
        Basis in which to express the mesh

    Returns
    -------
    XX, YY, ZZ : `~numpy.ndarray`
    """
    # Build coordinate array row-wise
    changed = np.empty(shape=(3, xx.size), dtype=float)
    linearized = np.empty(shape=(3, xx.size), dtype=float)
    linearized[0, :] = xx.ravel()
    linearized[1, :] = yy.ravel()
    linearized[2, :] = zz.ravel()

    # Change the basis at each row
    COB = change_of_basis(basis1, basis2)
    np.dot(COB, linearized, out=changed)
    return (
        changed[0, :].reshape(xx.shape),
        changed[1, :].reshape(yy.shape),
        changed[2, :].reshape(zz.shape),
    )


def minimum_image_distance(
    xx: np.ndarray, yy: np.ndarray, zz: np.ndarray, lattice: Basis
) -> np.ndarray:
    """
    Returns a periodic array according to the minimum image convention.

    Parameters
    ----------
    xx, yy, zz : ndarrays
        Arrays of equal shape, such as produced by numpy.meshgrid.
    lattice : list of ndarrays, shape(3,)
        Basis of the mesh

    Returns
    -------
    r : `~numpy.ndarray`
            Minimum image distance over the lattice
    """
    COB = change_of_basis(np.eye(3), lattice)
    linearized = np.empty(shape=(3, xx.size), dtype=float)  # In the standard basis
    ulinearized = np.empty_like(linearized)  # In the unitcell basis

    linearized[0, :] = xx.ravel()
    linearized[1, :] = yy.ravel()
    linearized[2, :] = zz.ravel()

    # Go to unitcell basis, where the cell is cubic of side length 1
    np.dot(COB, linearized, out=ulinearized)
    ulinearized -= np.rint(ulinearized)
    np.dot(np.linalg.inv(COB), ulinearized, out=linearized)

    return np.reshape(np.linalg.norm(linearized, axis=0), xx.shape)



# -*- coding: utf-8 -*-
import os
import socket
import tempfile
import pytest
from collections import Counter, namedtuple
from contextlib import suppress
from functools import lru_cache
from itertools import chain
from pathlib import Path
from tempfile import gettempdir
from warnings import catch_warnings, filterwarnings

import numpy as np
from crystals import CIFParser, Crystal, MPJParser, PDBParser, is_element
from crystals.parsers import STRUCTURE_CACHE, PWSCFParser

try:
    import Bio.PDB as biopdb
except ImportError:
    WITH_BIOPYTHON = False
else:
    WITH_BIOPYTHON = True

# API key to test Materials Project-related things
MPJ_API_KEY = os.environ.get("MATERIALS_PROJECT_API_KEY", None)

CIF_FILES = [Crystal.from_database(name).source for name in Crystal.builtins]
PWSCF_FILES = [
    Path(__file__).parent / "data" / stem
    for stem in ("pwscf_tise2.out", "pwscf_snse.out", "pwscf_graphite.out")
]

# Used to compare crystals.Atom instances and Bio.PDB.Atom instances
GenericAtom = namedtuple("GenericAtom", ["element", "coords"])

filterwarnings("ignore", category=UserWarning)


@lru_cache(maxsize=1)
def web_avail():
    """Returns whether or not an internet connection is available"""
    with suppress(OSError), socket.create_connection(("www.google.com", 80)):
        return True
    return False


@pytest.mark.skipif(not web_avail(), reason="Internet connection is required.")
def test_pdb_parser_fractional_atoms():
    """Test the PDBParser returns fractional atomic coordinates."""
    with tempfile.TemporaryDirectory() as temp_dir:
        with PDBParser("1fbb", download_dir=temp_dir) as parser:
            for atm in parser.atoms():
                pass  # TODO: find a test for this.


@pytest.mark.skipif(not web_avail(), reason="Internet connection is required.")
def test_pdb_parser_symmetry_operators():
    """Test that the non-translation part of the symmetry_operators is an invertible
    matrix of determinant 1 | -1"""
    with tempfile.TemporaryDirectory() as temp_dir:
        with PDBParser("1fbb", download_dir=temp_dir) as parser:
            for sym_op in parser.symmetry_operators():
                t = sym_op[:3, :3]
                assert round(abs(abs(np.linalg.det(t)) - 1), 5) == 0


@pytest.mark.skipif(not web_avail(), reason="Internet connection is required.")
def test_pdb_parser_residues():
    """Test the parsing of residues for 1fbb"""
    with tempfile.TemporaryDirectory() as temp_dir:
        with PDBParser("1fbb", download_dir=temp_dir) as parser:
            residues = list(parser.residues())
            atoms = list(parser.atoms())

    # Flatten residues into a list of atoms
    residue_atoms = list(chain.from_iterable(residues))
    assert len(residue_atoms) == len(atoms)


@pytest.mark.skipif(not web_avail(), reason="Internet connection is required.")
def test_pdb_parser_default_download_dir():
    """Test that the file is saved in the correct temporary directory by default"""
    filename = PDBParser.download_pdb_file("1fbb")

    assert filename.exists()
    assert filename.parent == STRUCTURE_CACHE


@pytest.mark.skipif(not WITH_BIOPYTHON, reason="Biopython is not installed/importable.")
@pytest.mark.skipif(not web_avail(), reason="Internet connection is required.")
@pytest.mark.parametrize("protein_id", ("1fbb", "1fat", "1gzx"))
def test_chemical_composition(protein_id):
    """Test crystals.PDBParser returns the same chemical composition as BIO.PDB.PDBParser implementation,
    i.e. the same elements in the right proportions."""
    pdb_list = biopdb.PDBList(verbose=False, obsolete_pdb=gettempdir())
    biopdb_parser = biopdb.PDBParser()

    with catch_warnings():
        filterwarnings("ignore", category=biopdb.PDBExceptions.PDBConstructionWarning)
        with tempfile.TemporaryDirectory() as temp_dir:
            with PDBParser(protein_id, download_dir=temp_dir) as parser:
                fname = pdb_list.retrieve_pdb_file(
                    protein_id, pdir=temp_dir, file_format="pdb"
                )

                # Note: Bio.PDB atoms store element as uppercase strings. Thus, they must be changed to titlecase
                crystals_chemical_composition = Counter(
                    [atm.element for atm in parser.atoms()]
                )
                biopdb_chemical_composition = Counter(
                    [
                        atm.element.title()
                        for atm in biopdb_parser.get_structure(
                            protein_id, fname
                        ).get_atoms()
                    ]
                )

                assert biopdb_chemical_composition == crystals_chemical_composition


@pytest.mark.parametrize("file", CIF_FILES)
def test_cif_compatibility(file):
    """Test the CIFParser on all CIF files stored herein to check build errors"""
    Crystal.from_cif(file)


def test_cif_uncertainties_issue7():
    """Test that the CIF parser handles atomic positions with uncertainties (issue #7)."""
    path = Path(__file__).parent / "data" / "issue7.cif"
    Crystal.from_cif(path)


@pytest.mark.parametrize("file", CIF_FILES)
def test_cif_fractional_atoms(file):
    """Test the CIFParser returns fractional atomic coordinates."""
    with CIFParser(file) as p:
        for atm in p.atoms():
            assert atm.coords_fractional.max() <= 1
            assert atm.coords_fractional.min() >= 0


@pytest.mark.parametrize("file", CIF_FILES)
def test_cif_symmetry_operators(file):
    """Test that the non-translation part of the symmetry_operators is an invertible
    matrix of determinant 1 | -1"""
    with CIFParser(file) as p:
        for sym_op in p.symmetry_operators():
            t = sym_op[:3, :3]
            assert round(abs(abs(np.linalg.det(t)) - 1), 7) == 0


def test_cif_silicon():
    """Test CIFParser on Si.cif (diamond structure)"""
    Si_path = Path(__file__).parent.parent / "cifs" / "Si.cif"
    si = Crystal.from_cif(Si_path)

    assert len(si) == 8


def test_cif_issue_5():
    """Test that the parsing of MgSiO3 is working. See Issue 5 for details on why this was a problem."""
    c = Crystal.from_cif(Path(__file__).parent / "data" / "issue5_MgSiO3.cif")
    assert len(c) == 20


def test_cif_vo2():
    """Test CIFParser on vo2.cif (monoclinic M1)"""
    VO2_path = Path(__file__).parent.parent / "cifs" / "vo2-m1.cif"
    vo2 = Crystal.from_cif(VO2_path)

    assert len(vo2) == 12
    assert np.allclose(vo2.lattice_parameters, (5.743, 4.517, 5.375, 90.0, 122.6, 90.0))
    assert round(abs(vo2.volume - 117.466_153_0), 7) == 0  # from cif2cell


def test_cif_site_occupancy():
    """Test that atom site occupancy is correctly parsed from CIF files."""
    path = Path(__file__).parent / "data" / "SiC_partial_site_occ.cif"
    with CIFParser(path) as parser:
        atoms = list(parser.atoms())

    for atm in filter(is_element("Si"), atoms):
        assert atm.occupancy == 0.75
    for atm in filter(is_element("C"), atoms):
        assert atm.occupancy == 0.85


@pytest.mark.xfail(reason="Bad SSL certificates")
@pytest.mark.skipif(
    MPJ_API_KEY is None,
    reason="Materials Project API key not defined in the environment.",
)
@pytest.mark.skipif(not web_avail(), reason="Internet connection is required.")
def test_mpj_example():
    """Test that the API example given on the Materials Project website is working as expected."""
    with tempfile.TemporaryDirectory() as temp_dir:
        cryst = Crystal.from_mp(
            api_key=MPJ_API_KEY,
            query="Fe2O3",
            download_dir=temp_dir,
            overwrite=True,
        )
        assert isinstance(cryst, Crystal)


@pytest.mark.xfail(reason="Bad SSL certificates")
@pytest.mark.skipif(
    MPJ_API_KEY is None,
    reason="Materials Project API key not defined in the environment.",
)
@pytest.mark.skipif(not web_avail(), reason="Internet connection is required.")
def test_material_id():
    """Test that that material ID for Fe2O3 is as expected."""
    with tempfile.TemporaryDirectory() as temp_dir:
        # We use the Zr5Te6 formula because it has a single Materials ID
        with MPJParser(
            api_key=MPJ_API_KEY,
            query="Zr5Te6",
            download_dir=temp_dir,
            overwrite=True,
        ) as parser:
            assert parser.material_id in "mp-29957"


@pytest.mark.parametrize("file,alat", zip(PWSCF_FILES, [6.6764, 21.4862, 4.6563]))
def test_pwscf_alat(file, alat):
    """Test the parsing of the lattice parameter (alat)"""
    parser = PWSCFParser(file)

    assert parser.alat == alat

    assert round(abs(parser.alat - parser.celldm[1]), 4) == 0


@pytest.mark.parametrize("file,natoms", zip(PWSCF_FILES, [3, 8, 4]))
def test_pwscf_natoms(file, natoms):
    """Test the parsing of the number of unit cell atoms"""
    parser = PWSCFParser(file)
    assert parser.natoms == natoms


def test_pwscf_lattice_vectors_alat():
    """Test the parsing of lattice vectors in alat units"""
    parser_tise2 = PWSCFParser(Path(__file__).parent / "data" / "pwscf_tise2.out")
    a1, a2, a3 = parser_tise2.lattice_vectors_alat()
    assert np.allclose(a1, np.array([0.989_891, -0.001_560, -0.001_990]))
    assert np.allclose(a2, np.array([-0.496_296, 0.856_491, 0.001_990]))
    assert np.allclose(a3, np.array([-0.003_315, 0.001_914, 1.669_621]))


def test_pwscf_reciprocal_vectors_alat():
    """Test the parsing of reciprocal vectors in alat units"""
    parser_tise2 = PWSCFParser(Path(__file__).parent / "data" / "pwscf_tise2.out")
    b1, b2, b3 = parser_tise2.reciprocal_vectors_alat()
    assert np.allclose(b1, np.array([1.011_138, 0.585_904, 0.001_336]))
    assert np.allclose(b2, np.array([0.001_839, 1.168_623, -0.001_336]))
    assert np.allclose(b3, np.array([0.001_203, -0.000_695, 0.598_942]))


@pytest.mark.parametrize("file", PWSCF_FILES)
def test_pwscf_atoms(file):
    parser = PWSCFParser(file)
    atoms = parser.atoms()
    assert len(atoms) == parser.natoms


@pytest.mark.parametrize("file,expected_spg", zip(PWSCF_FILES, [164, 62, 194]))
def test_pwscf_crystal_instance(file, expected_spg):
    """Test the construction of Crystal instances, and check against expected symmetry properties"""

    parser = PWSCFParser(file)
    crystal = Crystal.from_pwscf(file)

    # COmparison of lattice vectors
    assert np.allclose(
        np.array(crystal.lattice_vectors),
        np.array(parser.lattice_vectors()),
    )

    # Comparison requires slighly relaxed precision
    assert crystal.symmetry(symprec=1e-1)["international_number"] == expected_spg


@pytest.mark.parametrize(
    "poscar", ["issue5_MgSiO3.cartesian.POSCAR", "issue5_MgSiO3.direct.POSCAR"]
)
def test_pwscf_lattice_vectors_alat(poscar):
    """Test POSCAR parser against CIF parser"""
    c0 = Crystal.from_cif(Path(__file__).parent / "data" / "issue5_MgSiO3.cif")
    c1 = Crystal.from_poscar(Path(__file__).parent / "data" / poscar)
    assert c0 == c1



# -*- encoding: utf-8 -*-

import re
from glob import glob
from itertools import chain
from pathlib import Path

from setuptools import find_packages, setup

PACKAGE_NAME = "crystals"
DESCRIPTION = "Data structures for crystallography"
URL = "http://crystals.readthedocs.io"
DOWNLOAD_URL = "http://github.com/LaurentRDC/crystals"
AUTHOR = "Laurent P. René de Cotret"
AUTHOR_EMAIL = "laurent.decotret@outlook.com"

CIF_FILES = chain.from_iterable([glob("crystals\\cifs\\*.cif")])

base_path = Path(__file__).parent
with open(base_path / PACKAGE_NAME / "__init__.py") as f:
    module_content = f.read()
    VERSION = (
        re.compile(r".*__version__ = \"(.*?)\"", re.S).match(module_content).group(1)
    )
    LICENSE = (
        re.compile(r".*__license__ = \"(.*?)\"", re.S).match(module_content).group(1)
    )

with open("README.md", encoding="utf-8") as f:
    README = f.read()

with open("requirements.txt") as f:
    REQUIREMENTS = [line for line in f.read().split("\n") if len(line.strip())]


if __name__ == "__main__":
    setup(
        name=PACKAGE_NAME,
        description=DESCRIPTION,
        long_description=README,
        long_description_content_type="text/markdown",
        license=LICENSE,
        url=URL,
        download_url=DOWNLOAD_URL,
        version=VERSION,
        author=AUTHOR,
        author_email=AUTHOR_EMAIL,
        maintainer=AUTHOR,
        maintainer_email=AUTHOR_EMAIL,
        install_requires=REQUIREMENTS,
        keywords=["crystallography", "material science", "structural biology"],
        project_urls={
            "Documentation": "https://crystals.readthedocs.io/",
            "Source": "https://github.com/LaurentRDC/crystals",
        },
        python_requires=">=3.7",
        packages=find_packages(),
        data_files=[("crystals\\cifs", CIF_FILES)],
        include_package_data=True,
        zip_safe=False,
        entry_points={"console_scripts": ["crystals = crystals.__main__:main"]},
        # list of possible classifiers:
        #  https://pypi.python.org/pypi?%3Aaction=list_classifiers
        classifiers=[
            "Environment :: Console",
            "Development Status :: 5 - Production/Stable",
            "Intended Audience :: Developers",
            "Intended Audience :: Education",
            "Intended Audience :: Science/Research",
            "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
            "Natural Language :: English",
            "Operating System :: OS Independent",
            "Programming Language :: Python",
            "Programming Language :: Python :: 3",
            "Topic :: Scientific/Engineering",
            "Topic :: Scientific/Engineering :: Chemistry",
            "Topic :: Scientific/Engineering :: Physics",
            "Topic :: Scientific/Engineering :: Bio-Informatics",
        ],
    )



# -*- coding: utf-8 -*-
"""
Conversion between ``crystals`` data structures and other modules.

These functions are not expected to be used on their own; see the associated
 `Crystal` methods instead, like `Crystal.to_cif`.
"""
from contextlib import redirect_stdout
from io import StringIO
from itertools import groupby
from os import PathLike
from typing import Iterable, Optional

import numpy as np
from CifFile import CifBlock, CifFile

from . import __version__
from .atom import Atom

try:
    import ase
except ImportError:
    WITH_ASE = False
else:
    WITH_ASE = True

CIF_HEADER = f"""
#\\#CIF_2.0
#
# File generated by `crystals` {__version__}, documented at https://crystals.rtfd.io
# Reference:  L. P. René de Cotret et al, An open-source software ecosystem for the interactive exploration
#             of ultrafast electron scattering data, Advanced Structural and Chemical Imaging 4:11 (2018) DOI: 10.1186/s40679-018-0060-y.
#
# For more information on this type of file, please visit https://www.iucr.org/resources/cif
"""


# TODO: test against known XYZ file
def write_xyz(crystal: Iterable[Atom], fname: PathLike, comment: Optional[str] = None):
    """
    Generate an atomic coordinates .xyz file from a crystal structure.

    Parameters
    ----------
    crystal : crystals.Crystal
        Crystal to be converted.
    fname : path-like
        The XYZ file will be written to this file. If the file already exists,
        it will be overwritten.
    comment : str or None, optional
        Comment to include at the second line of ``fname``.
    """
    # Format is specified here:
    #   http://openbabel.org/wiki/XYZ_%28format%29
    comment = comment or ""

    with open(fname, "wt", encoding="ascii") as file:
        # First two lines are:
        #   1. Number of atoms described in the file
        #   2. Optional comment
        file.write(str(len(crystal)) + "\n")
        file.write(comment + "\n")

        # Write atomic data row-by-row
        # For easier human readability, atoms are sorted
        # by element
        for atom in sorted(crystal):
            x, y, z = atom.coords_cartesian
            row = f"  {atom.element:<2}       {x:10.5f}       {y:10.5f}       {z:10.5f}"
            file.write(row + "\n")


# TODO: write the asymmetric cell + symmetry operatrs
def write_cif(crystal: "Crystal", fname: PathLike):
    """
    Generate an atomic coordinates .cif file from a crystal structure.

    .. versionadded:: 1.2.0

    Parameters
    ----------
    crystal : crystals.Crystal
        Crystal to be converted.
    fname : path-like
        The CIF file will be written to this file. If the file already exists,
        it will be overwritten.
    comment : str or None, optional
        Comment to include at the second line of ``fname``.
    """
    cf = CifFile(strict=False)
    a, b, c, alpha, beta, gamma = crystal.lattice_parameters
    lattice_items = {
        "_cell_length_a": a,
        "_cell_length_b": b,
        "_cell_length_c": c,
        "_cell_angle_alpha": alpha,
        "_cell_angle_beta": beta,
        "_cell_angle_gamma": gamma,
    }

    sym = crystal.symmetry()
    symmetry_items = {
        "_symmetry_Int_Tables_number": sym["international_number"],
        "_symmetry_space_group_name_Hall": sym["hall_symbol"],
    }

    block = CifBlock()
    for key, val in lattice_items.items():
        block[key] = val

    for key, val in symmetry_items.items():
        block[key] = val

    # Note that we are using all atoms in the unit-cell,
    # and not the asymmetric unit cell + symmetry operators
    # This is valid CIF! And it is much simpler to implement
    # TODO: how to determine asymmetric cell + symmetry operations?
    atoms = list(crystal.unitcell)
    symbols = [atm.symbol for atm in atoms]
    xf = [atm.coords_fractional[0] for atm in atoms]
    yf = [atm.coords_fractional[1] for atm in atoms]
    zf = [atm.coords_fractional[2] for atm in atoms]

    block.CreateLoop(
        datanames=[
            "_atom_site_type_symbol",
            "_atom_site_fract_x",
            "_atom_site_fract_y",
            "_atom_site_fract_z",
        ],
        length_check=False,
    )
    block["_atom_site_type_symbol"] = symbols
    block["_atom_site_fract_x"] = xf
    block["_atom_site_fract_y"] = yf
    block["_atom_site_fract_z"] = zf

    # Name of the block cannot be empty!
    block_name = crystal.chemical_formula.replace(" ", "_")
    cf[block_name] = block

    # Converting to string writes to stdout for some reason
    with redirect_stdout(StringIO()):
        lines = str(cf).splitlines()

    with open(fname, "w", encoding="utf-8") as f:
        f.write(CIF_HEADER)
        f.write("\n".join(lines[13::]))  # Skip the fixed header


def ase_atoms(crystal: "Crystal", **kwargs):
    """
    Convert a :class:`crystals.Crystal` object into an :class:`ase.Atoms` object.
    Keyword arguments are passed to :class:`ase.Atoms` constructor.

    Parameters
    ----------
    crystal : :class:`Crystal`
        Crystal to be converted.

    Returns
    -------
    atoms : ase.Atoms
        Group of atoms ready for ASE's routines.

    Raises
    ------
    ImportError : If ASE is not installed
    """
    if not WITH_ASE:
        raise ImportError("ASE is not installed/importable.")

    return ase.Atoms(
        symbols=[
            ase.Atom(
                symbol=atom.element,
                position=atom.coords_cartesian,
                magmom=atom.magmom,
                mass=atom.mass,
            )
            for atom in crystal
        ],
        cell=np.array(crystal.lattice_vectors),
        **kwargs,
    )


def write_poscar(
    crystal: "Crystal",
    fname: PathLike,
    comment: Optional[str] = None,
    scaling_factor: float = 1.0,
    cartesian: bool = False,
):
    """
    Generate a POSCAR file from a crystal structure.

    Parameters
    ----------
    crystal : crystals.Crystal
        Crystal to be converted.
    fname : path-like
        The POSCAR file will be written to this file. If the file already exists,
        it will be overwritten.
    comment : str or None, optional
        Comment to include as the first line of ``fname``.
    scaling_factor: float, optional
        Scaling factor with which to scale lattice vectors
    cartesian: bool, optional
        If True, the resulting file will be expressed in cartesian coordinates.
        Default is to use fractional coordinates.
    """
    # Format is specified here:
    #   https://www.vasp.at/wiki/index.php/POSCAR
    comment = comment or "POSCAR"

    with open(fname, "wt", encoding="ascii") as file:
        file.write(comment + "\n")

        file.write(str(scaling_factor))
        file.write("\n")
        for vec in crystal.lattice_vectors:
            x, y, z = vec / scaling_factor
            file.write(f"{x:10.6f}  {y:10.6f}  {z:10.6f}")
            file.write("\n")

        grouped = [
            (elem, list(atoms))
            for elem, atoms in groupby(sorted(crystal), key=lambda atom: atom.element)
        ]

        file.write(" ".join(["%5s" % elem for elem, atoms in grouped]))
        file.write("\n")
        file.write(" ".join(["%5d" % len(atoms) for elem, atoms in grouped]))
        file.write("\n")

        if cartesian:
            file.write("Cartesian\n")
            for elem, atoms in grouped:
                for atom in atoms:
                    x, y, z = atom.coords_cartesian / scaling_factor
                    row = f"{x:10.5f}  {y:10.5f}  {z:10.5f}"
                    file.write(row + "\n")
        else:
            file.write("Direct\n")
            for elem, atoms in grouped:
                for atom in atoms:
                    x, y, z = atom.coords_fractional
                    row = f"{x:10.5f}  {y:10.5f}  {z:10.5f}"
                    file.write(row + "\n")



# -*- coding: utf-8 -*-
from copy import deepcopy
from math import degrees, radians
from itertools import islice

import numpy as np
from numpy.linalg import norm

from crystals import Crystal, Lattice, LatticeSystem
from crystals.lattice import lattice_vectors_from_parameters
from crystals.affine import rotation_matrix
import pytest

np.random.seed(23)


def angle_between(v1, v2):
    """Returns the angle in degrees between vectors `v1` and `v2`"""
    v1 /= norm(v1)
    v2 /= norm(v2)
    rad = np.arccos(np.clip(np.dot(v1, v2), -1.0, 1.0))
    return degrees(rad)


def test_euclidian_lattice():
    lattice = Lattice(np.eye(3))
    assert np.array_equal(lattice.a1, [1, 0, 0])
    assert np.array_equal(lattice.a2, [0, 1, 0])
    assert np.array_equal(lattice.a3, [0, 0, 1])

    assert lattice.volume == 1


def test_euclidian_lattice_equality():
    """Test equality between identical Lattice instances and copies"""
    lattice = Lattice(np.eye(3))
    assert lattice == lattice
    assert lattice == deepcopy(lattice)

    assert lattice != Lattice(2 * np.eye(3))


def test_lattice_array_shape():
    """Test that array(Lattice(...)) is always 3x3"""
    arr = np.random.random(size=(3, 3))
    lattice = Lattice(arr)
    assert np.allclose(arr, np.array(lattice))


def test_lattice_array_dtype():
    """Test that the data-type of array(Lattice(...)) is respected"""
    arr = np.random.random(size=(3, 3))
    lattice = Lattice(arr)
    assert np.array(lattice, dtype=int).dtype, int


def test_lattice_frac_mesh():
    """Test that Lattice.frac_mesh is working as expected compared to numpy.meshgrid"""
    lattice = Lattice(np.eye(3))
    x = np.linspace(0, 1, num=8)

    for out, n in zip(lattice.frac_mesh(x), np.meshgrid(x, x, x)):
        assert np.allclose(out, n)


def test_lattice_frac_mesh_two_arr():
    """Test that Lattice.frac_mesh is raising an exception for two input vectors"""
    lattice = Lattice(np.eye(3))
    x = np.linspace(0, 1, num=2)

    with pytest.raises(ValueError):
        lattice.frac_mesh(x, x)


def test_lattice_real_mesh_trivial():
    """Test that Lattice.mesh works identically to Lattice.frac_mesh for trivial lattice"""
    lattice = Lattice(np.eye(3))
    x = np.linspace(0, 1, num=8)

    for frac, real in zip(lattice.frac_mesh(x), lattice.mesh(x)):
        assert np.allclose(frac, real)


def test_lattice_real_mesh():
    """Test that Lattice.mesh works as expected"""
    lattice = Lattice(2 * np.eye(3))
    x = np.linspace(0, 1, num=8)

    # since lattice is a stretched euclidian lattice, we expect
    # a maximum length of 2
    assert np.max(lattice.mesh(x)[0]) == 2


def test_lattice_parameters_orthorombic():
    """alpha = beta = gamma = 90"""
    a1, a2, a3 = Lattice.from_parameters(2, 1, 5, 90, 90, 90).lattice_vectors
    assert np.allclose(a1, [2, 0, 0])
    assert np.allclose(a2, [0, 1, 0])
    assert np.allclose(a3, [0, 0, 5])


def test_lattice_parameters_monoclinic():
    """beta =/= 90"""
    a1, a2, a3 = Lattice.from_parameters(1, 2, 3, 90, 120, 90).lattice_vectors

    assert round(abs(norm(a1) - 1), 7) == 0
    assert round(abs(norm(a2) - 2), 7) == 0
    assert round(abs(norm(a3) - 3), 7) == 0

    assert round(abs(angle_between(a2, a3) - 90), 7) == 0
    assert round(abs(angle_between(a3, a1) - 120), 7) == 0
    assert round(abs(angle_between(a1, a2) - 90), 7) == 0


def test_lattice_parameters_triclinic():
    """alpha, beta, gama =/= 90"""
    a1, a2, a3 = Lattice.from_parameters(1, 2, 3, 75, 40, 81).lattice_vectors

    assert round(abs(norm(a1) - 1), 7) == 0
    assert round(abs(norm(a2) - 2), 7) == 0
    assert round(abs(norm(a3) - 3), 7) == 0

    assert round(abs(angle_between(a2, a3) - 75), 7) == 0
    assert round(abs(angle_between(a3, a1) - 40), 7) == 0
    assert round(abs(angle_between(a1, a2) - 81), 7) == 0


def test_lattice_parameters_reciprocal_and_back():
    """Create lattice from parameters, take reciprocal twice,
    and see if the parameters have changed."""
    triclinic = (3, 4, 20, 45, 90, 126)
    triclinic2 = Lattice.from_parameters(
        *triclinic
    ).reciprocal.reciprocal.lattice_parameters
    assert np.allclose(triclinic, triclinic2)


@pytest.mark.parametrize("name", islice(Crystal.builtins, 20))
def test_lattice_parameters_back_and_forth(name):
    """Test that the conversion between lattice vectors and lattice parameters
    is working"""

    c = Crystal.from_database(name)
    lv1 = c.lattice_vectors
    params = c.lattice_parameters
    lv2 = lattice_vectors_from_parameters(*params)

    assert np.allclose(lv1, lv2)


def test_cubic_lattice_system():
    """Test that the Lattice.lattice_system attribute is working properly for cubic lattice"""
    assert Lattice(2 * np.eye(3)).lattice_system == LatticeSystem.cubic


def test_tetragonal_lattice_system():
    """Test that the Lattice.lattice_system attribute is working properly for tetragonal lattice"""
    parameters = (2, 2, 3, 90, 90, 90)
    l = Lattice.from_parameters(*parameters)
    assert l.lattice_system == LatticeSystem.tetragonal


def test_rhombohedral_lattice_system():
    """Test that the Lattice.lattice_system attribute is working properly for rhombohedral lattice"""
    parameters = (1, 1, 1, 87, 87, 87)
    l = Lattice.from_parameters(*parameters)
    assert l.lattice_system == LatticeSystem.rhombohedral


def test_monoclinic_lattice_system():
    """Test that the Lattice.lattice_system attribute is working properly for monoclinic lattice
    including all possible permutations."""
    parameters = (1, 2, 3, 90, 115, 90)
    l = Lattice.from_parameters(*parameters)
    assert l.lattice_system == LatticeSystem.monoclinic

    parameters = (2, 3, 1, 115, 90, 90)
    l = Lattice.from_parameters(*parameters)
    assert l.lattice_system == LatticeSystem.monoclinic

    parameters = (3, 1, 2, 90, 90, 115)
    l = Lattice.from_parameters(*parameters)
    assert l.lattice_system == LatticeSystem.monoclinic


def test_hexagonal_lattice_system():
    """Test that the Lattice.lattice_system attribute is working properly for hexagonal lattice,
    including all possible permutations of lattice parameters."""
    parameters = (2, 2, 3, 90, 90, 120)
    l = Lattice.from_parameters(*parameters)
    assert l.lattice_system == LatticeSystem.hexagonal

    parameters = (3, 2, 2, 120, 90, 90)
    l = Lattice.from_parameters(*parameters)
    assert l.lattice_system == LatticeSystem.hexagonal

    parameters = (2, 3, 2, 90, 120, 90)
    l = Lattice.from_parameters(*parameters)
    assert l.lattice_system == LatticeSystem.hexagonal

    parameters = (2, 2, 2, 90, 120, 90)
    l = Lattice.from_parameters(*parameters)
    assert l.lattice_system == LatticeSystem.hexagonal


def test_triclinic_lattice_system():
    """Test that the Lattice.lattice_system attribute is working properly for triclinic lattice"""
    l = Lattice.from_parameters(1, 2, 3, 75, 40, 81)
    assert l.lattice_system == LatticeSystem.triclinic


def test_lattice_system_graphite():
    """Test that the builtin Crystal for graphite has a hexagonal lattice system"""
    graphite = Crystal.from_database("C")
    assert graphite.lattice_system == LatticeSystem.hexagonal


def test_lattice_system_lead():
    """Test that the builtin Crystal for lead has a cubic lattice system"""
    pb = Crystal.from_database("Pb")
    assert pb.lattice_system == LatticeSystem.cubic


def test_lattice_system_vo2():
    """Test that the builtin Crystal for monoclinic M1 VO2 has a monoclinic lattice system"""
    vo2 = Crystal.from_database("vo2-m1")
    assert vo2.lattice_system == LatticeSystem.monoclinic


def test_scattering_vector_trivial():
    """Test that Lattice.scattering_vectors is working"""
    lattice = Lattice(np.random.random((3, 3)))
    assert np.allclose(lattice.scattering_vector((0, 0, 0)), ((0, 0, 0)))


def test_scattering_vector_table():
    """Test that Lattice.scattering_vector is working on tables of reflections"""
    lattice = Lattice(2 * np.pi * np.eye(3))
    vectors = np.random.random(size=(10, 3))
    assert np.allclose(lattice.scattering_vector(vectors), vectors)


def test_miller_indices_trivial():
    """Test that Lattice.miller_indices is working"""
    lattice = Lattice(np.random.random((3, 3)))
    assert np.allclose(lattice.miller_indices((0, 0, 0)), ((0, 0, 0)))


def test_miller_indices_table():
    """Test that Lattice.miller_indices is working on tables of vectors"""
    lattice = Lattice(2 * np.pi * np.eye(3))
    vectors = np.random.random(size=(10, 3))
    assert np.allclose(lattice.miller_indices(vectors), vectors)


def test_back_and_forth():
    """Test that Lattice.miller_indices and Lattice.scattering_vector are
    reciprocal to each other"""
    lattice = Lattice(np.random.random((3, 3)))
    h, k, l = np.random.randint(-10, 10, size=(3,))
    vector = lattice.scattering_vector((h, k, l))
    hp, kp, lp = lattice.miller_indices(vector)

    assert round(abs(h - float(hp)), 7) == 0
    assert round(abs(k - float(kp)), 7) == 0
    assert round(abs(l - float(lp)), 7) == 0


def test_bounded_reflections_negative():
    """Test that negative reflection bounds raise an Exception.
    Otherwise, an infinite number of reflections will be generated"""
    crystal = Crystal.from_database("C")
    with pytest.raises(ValueError):
        next(crystal.bounded_reflections(-1))


def test_bounded_reflections_zero():
    """Check that bounded_reflections returns (000) for a zero bound"""
    crystal = Crystal.from_database("C")
    reflections = set(crystal.bounded_reflections(0))
    assert (0, 0, 0) in reflections
    assert len(reflections), 1


@pytest.mark.parametrize(
    "crystal", map(Crystal.from_database, islice(Crystal.builtins, 3))
)
def test_bounded_reflections_all_within_bounds(crystal):
    """Check that every reflection is within the bound"""
    bound = 10
    vectors = (
        crystal.scattering_vector(refl) for refl in crystal.bounded_reflections(bound)
    )
    for vector in vectors:
        assert np.linalg.norm(vector) <= bound



# -*- coding: utf-8 -*-

from collections import Counter, OrderedDict
from copy import deepcopy
from functools import reduce
from itertools import chain
from math import gcd
from operator import itemgetter
from typing import Any, Callable, Dict, Iterable, Iterator, Optional, Union

import numpy as np
from numpy.typing import ArrayLike

from .affine import affine_map
from .atom import Atom


class AtomicStructure:
    """
    Base class for atomic structures. These structures can be made
    out of :class:`Atom` objects, or other :class:`AtomicStructures` subclasses.

    The :class:`AtomicStructures` class provides an abstraction over structure with and without
    substructure. Subclasses can be iterated over as an iterable of atoms. Order of iteration
    is not guaranteed.

    Hierarchical containership of :class:`AtomicStructures` is implemented, as well as
    containership of :class:`Atom` instances.

    Parameters
    ----------
    atoms : iterable, optional
        Atoms not attached to a substructure.
    substructures : iterable, optional
        Atomic substructures. For example, different atom chains could form
        a secondary structure in a protein.
    """

    def __init__(
        self,
        atoms: Optional[Iterable[Atom]] = None,
        substructures: Optional[Iterable["AtomicStructure"]] = None,
        **kwargs,
    ):
        self.atoms = frozenset(atoms or {})
        self.substructures = frozenset(substructures or {})
        super().__init__(**kwargs)

    def __iter__(self) -> Iterator[Atom]:
        """Yields :class:`Atom` instances from the structure and substructures
        recursively. Order is not guaranteed."""
        yield from iter(self.atoms)
        yield from chain.from_iterable(self.substructures)

    def __contains__(self, item: Union[Atom, "AtomicStructure"]) -> bool:
        """Check containership of :class:`Atom` instances or :class:`AtomicStructure` substructures recursively."""
        if isinstance(item, AtomicStructure):
            return item in self.substructures

        # Either the item is an orphan atom or
        # it is in one of the substructures
        # Checking containership of sets is faster than iterating
        return any(
            (item in struct) for struct in chain([self.atoms], self.substructures)
        )

    def __len__(self) -> int:
        """Number of :class:`Atom` instances present in the structure and substructures"""
        return len(self.atoms) + sum(len(struct) for struct in self.substructures)

    def __bool__(self) -> bool:
        """An :class:`AtomicStructure` is `False` if empty, and `True` otherwise"""
        return bool(self.atoms) or bool(self.substructures)

    def __add__(self, other: "AtomicStructure") -> "AtomicStructure":
        """Create a new structure made from the combination of two structures."""
        # We defer construction to the current class. Therefore, subclasses of AtomicStructure
        # will add into their own class
        return self.__class__(
            atoms=deepcopy(self.atoms) | deepcopy(other.atoms),
            substructures=deepcopy(self.substructures) | deepcopy(other.substructures),
        )

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, AtomicStructure):
            return (self.atoms == other.atoms) and (
                self.substructures == other.substructures
            )
        return NotImplemented

    def __hash__(self) -> int:
        return hash(self.atoms) | hash(self.substructures)

    def __repr__(self) -> str:
        """Verbose string representation of this instance."""
        # AtomicStructure subclasses need not override this method
        # since the class name is dynamically determined
        rep = f"< {self.__class__.__name__} object with following orphan atoms:"

        # Note that repr(Atom(...)) includes these '< ... >'
        # We remove those for cleaner string representation
        # Sorting the atoms is also important, so that the expected outputs
        # in doctests don't change.
        for atm in sorted(self):
            rep += "\n    " + repr(atm).replace("<", "").replace(">", "").strip()

        if self.substructures:
            rep += "and the following substructures:"
            for struct in self.substructures:
                rep += "\n" + repr(struct)

        return rep + " >"

    def __array__(self, *args, **kwargs) -> np.ndarray:
        """
        Returns an array in which each row represents an :class:`Atom` instance.
        Atoms are ordered by atomic number.
        """
        # Not all keyword arguments can be passed to 'np.empty'
        kwargs.pop("copy", None)
        arr = np.empty(shape=(len(self), 4), *args, **kwargs)
        atoms = sorted(self)
        for row, atm in enumerate(atoms):
            arr[row, :] = np.array(atm, *args, **kwargs)
        return arr

    def itersorted(
        self, *, key: Callable[[Atom], Any] = None, reverse: bool = False
    ) -> Iterator[Atom]:
        """
        Yields :class:`Atom` in sorted order. By default, atoms are sorted by element.

        Parameters
        ----------
        key : callable or None, optional
            Function of one argument that is used to extract a comparison key for an :class:`Atom` instance.
        reverse : bool, optional
            If True, elements are yielded in reverse order.

        Yields
        ------
        atm : `Atom`
        """
        # TODO: deprecate
        yield from sorted(iter(self), key=key, reverse=reverse)

    def satisfying(self, predicate: Callable[[Atom], bool]) -> "AtomicStructure":
        """
        Builds a new AtomicStructure from atoms satisfying a predicate.

        This method is similar to the built-in filter, except that the properties
        of AtomicStructures are conserved.

        Parameters
        ----------
        predicate : callable
            Function that takes a single atom, and returns boolean.

        Returns
        -------
        out : AtomicStructure
        """
        return AtomicStructure(atoms=filter(predicate, iter(self)))

    @property
    def chemical_composition(self) -> Dict[str, float]:
        """
        Chemical composition of this structure as an ordered dictionary. Keys are elemental symbols.
        Elements are in descending order of prevalence.

        Example
        -------
        >>> c = Crystal.from_cod(1508819)
        >>> c.chemical_composition # doctest: +SKIP
        OrderedDict([('H',  0.441),
                     ('C',  0.349),
                     ('N',  0.093),
                     ('O',  0.093),
                     ('Cl', 0.023)
                     ])
        """
        # We can't use a Counter directly since Counter values are integer by default
        number_atoms = len(self)
        counter = Counter(atm.element for atm in self)
        # For display reasons, it makes more sense to sort elements in descending percentage
        # i.e. more prevalent elemts are inserted in the dictionary first
        sorted_by_percentage = sorted(counter.items(), key=itemgetter(-1), reverse=True)
        return OrderedDict((k, v / number_atoms) for k, v in sorted_by_percentage)

    @property
    def chemical_formula(self) -> str:
        """
        Empirical chemical formula for this structure based on the chemical symbols.
        The string is returned in Hill notation: symbols are alphabetically ordered
        except for carbon (C) and hydrogen (H), which are put first.

        Example
        -------
        >>> c = Crystal.from_cod(1508819)
        >>> c.chemical_formula
        'C15 H19 Cl N4 O4'
        """
        symbols_count = Counter(atm.element for atm in self)

        # Special case: only one distinct element
        if len(symbols_count) == 1:
            return next(iter(symbols_count.keys()))

        # Calculate greatest common divisor for all elements
        # Note that we are guaranteed to have at least two distinct elements
        counts = list(symbols_count.values())
        common_divisor = reduce(gcd, counts[1:], counts[0])
        symbols_count_empirical = {
            k: int(v / common_divisor) for k, v in symbols_count.items()
        }

        # Build a list of elements and their counts
        # We put carbon and hydrogen first because convention
        elements = []
        for elem in ("C", "H"):
            if elem in symbols_count_empirical:
                elements.append((elem, symbols_count_empirical.pop(elem)))

        # Then, elements are in alphabetical number
        elements.extend(
            [(e, symbols_count_empirical[e]) for e in sorted(symbols_count_empirical)]
        )
        return " ".join(
            f"{symbol}{count}" if count > 1 else symbol for symbol, count in elements
        )

    def transform(self, *operators: ArrayLike) -> "AtomicStructure":
        """
        Return a transformed AtomicStructure based on symmetry operators.

        Parameters
        ----------
        operators : iterable of array_like
            Symmetry operators, either 3x3 or 4x4.

        Returns
        -------
        transformed : AtomicStructure
            New structure.
        """
        operators = tuple(map(affine_map, operators))

        transformed_atoms = set()
        for atom in self.atoms:
            for operator in operators:
                transformed_atoms.add(atom.transform(operator))

        # Substructures are recursively transformed
        transformed_substructures = set()
        for substructure in self.substructures:
            transformed_substructures.add(substructure.transform(*operators))

        # We defer construction to the current class. Therefore, subclasses of AtomicStructure
        # will transform into their own class
        return self.__class__(
            atoms=transformed_atoms, substructures=transformed_substructures
        )
